@@ -40,12 +40,23 @@
 ;; Definition in assembler (see CONTROL.Q)
 |#
 
-(defun macro-call-error (macro-form)
+(defun macro-call-error (macro-form) ; ABI
   (error-of-type 'source-program-error
+    :form macro-form
+    :detail macro-form
     (TEXT "The macro ~S may not be called with ~S arguments: ~S")
     (car macro-form) (1- (length macro-form)) macro-form))
 
+(defun macro-nonnull-element-error (macro-form macro-name element) ; ABI
+  (error-of-type 'source-program-error
+    :form macro-form
+    :detail element
+    (TEXT "~S: ~S does not match lambda list element ~:S")
+    macro-name element '()))
+
 (proclaim '(special
+        %whole-form ;; the whole source form being macroexpanded or compiled
+
         %restp ;; indicates whether &REST/&BODY/&KEY was given,
                ;; and therefore the number of arguments is unbound.
 
@@ -61,6 +72,8 @@
         %default-form ;; default form for optional and keyword arguments,
                       ;; for which no default form is supplied
                       ;; NIL normally, or (QUOTE *) for DEFTYPE.
+
+        %null-tests   ;; tests for () matches
 ))
 #|
  (ANALYZE1 lambdalist accessexp name wholevar)
@@ -90,14 +103,14 @@ creates a testform from %restp, %min-args, %arg-count, which indicates
 during evaluation whether the variable value of the symbol can be a
 function call for the macro.
 
- (MAKE-MACRO-EXPANSION macrodef)
+ (MAKE-MACRO-EXPANSION macrodef whole-form)
 returns, for a macro definition macrodef = (name lambdalist . body),
 1. the macro-expander as the program text (FUNCTION ... (LAMBDA ..)),
 2. name, a symbol
 3. lambda list
 4. docstring (or NIL, if not there)
 
- (MAKE-MACRO-EXPANDER macrodef &optional env)
+ (MAKE-MACRO-EXPANDER macrodef whole-form &optional env)
 returns, for a macro definition macrodef = (name lambdalist . body),
 the actual object #<MACRO expander> for the FENV.
 |#
@@ -112,6 +125,8 @@ the actual object #<MACRO expander> for the FENV.
     (cond ((symbolp (car listr)) (setq %let-list (cons `(,(car listr) nil) %let-list)))
           ((atom (car listr))
            (error-of-type 'source-program-error
+             :form %whole-form
+             :detail (car listr)
              (TEXT "in macro ~S: ~S may not be used as &AUX variable.")
              name (car listr)))
           (t (setq %let-list
@@ -121,6 +136,8 @@ the actual object #<MACRO expander> for the FENV.
   (when (and (consp (cdr item)) (consp (cddr item)))
     (unless (symbolp (caddr item))
       (error-of-type 'source-program-error
+        :form %whole-form
+        :detail (caddr item)
         (TEXT "~S: invalid supplied-p variable ~S")
         name (caddr item)))
     (third item)))
@@ -129,7 +146,7 @@ the actual object #<MACRO expander> for the FENV.
 (defun kwd-arg-form (restvar kw svar default &aux (dummy '#.(cons nil nil)))
   ;; the default value should not be evaluated unless it is actually used
   (let ((arg (gensym "KWD-ARG-")))
-    `(let ((,arg (GETF ,restvar ,kw ',dummy)))
+    `(let ((,arg (GETF ,restvar ',kw ',dummy)))
        (if (eq ,arg ',dummy)
            (progn ,@(when svar `((setq ,svar nil))) ,default)
            ,arg))))
@@ -161,16 +178,16 @@ the actual object #<MACRO expander> for the FENV.
                    ((and (consp next) (eql (length next) 1))
                     (setq next (list (car next) %default-form)))))
            (cond ((symbolp next)  ; foo
-                  (setq kw (intern (symbol-name next) *keyword-package*))
+                  (setq kw (symbol-to-keyword next))
                   (setq %let-list
-                        (cons `(,next (GETF ,restvar ,kw NIL)) %let-list))
+                        (cons `(,next (GETF ,restvar ',kw NIL)) %let-list))
                   (setq kwlist (cons kw kwlist)))
                  ((atom next)
                   (cerror (TEXT "It will be ignored.")
                           (TEXT "The lambda list of macro ~S contains the invalid element ~S")
                           name next))
                  ((symbolp (car next))  ; (foo ...)
-                  (setq kw (intern (symbol-name (car next)) *keyword-package*))
+                  (setq kw (symbol-to-keyword (car next)))
                   (when (setq svar (get-supplied-p name next))
                     (setq %let-list (cons `(,svar t) %let-list)))
                   (setq %let-list
@@ -212,6 +229,8 @@ the actual object #<MACRO expander> for the FENV.
 (defun analyze-rest (lambdalistr restexp name)
   (if (atom lambdalistr)
       (error-of-type 'source-program-error
+        :form %whole-form
+        :detail lambdalistr
         (TEXT "The lambda list of macro ~S is missing a variable after &REST/&BODY.")
         name))
   (let ((restvar (car lambdalistr))
@@ -221,6 +240,8 @@ the actual object #<MACRO expander> for the FENV.
            (setq %let-list (cons `(,restvar ,restexp) %let-list)))
           ((atom restvar)
            (error-of-type 'source-program-error
+             :form %whole-form
+             :detail restvar
              (TEXT "The lambda list of macro ~S contains an illegal variable after &REST/&BODY: ~S")
              name restvar))
           (t
@@ -266,6 +287,13 @@ the actual object #<MACRO expander> for the FENV.
    (cons (cdr h) (cdr exp))
    (list 'cdr exp)))
 
+(defun empty-pattern (name accessexp wholevar)
+  (let ((g (gensym)))
+    (setq %let-list (cons `(,g ,(cons-car accessexp)) %let-list))
+    (setq %null-tests
+          (cons `(WHEN ,g (MACRO-NONNULL-ELEMENT-ERROR ,wholevar ',name ,g))
+                 %null-tests))))
+
 (defun analyze1 (lambdalist accessexp name wholevar)
   (do ((listr lambdalist (cdr listr))
        (within-optional nil)
@@ -275,6 +303,8 @@ the actual object #<MACRO expander> for the FENV.
        (when listr
          (unless (symbolp listr)
            (error-of-type 'source-program-error
+             :form %whole-form
+             :detail listr
              (TEXT "The lambda list of macro ~S contains an illegal &REST variable: ~S")
              name listr))
          (setq %let-list (cons `(,listr ,accessexp) %let-list))
@@ -289,6 +319,8 @@ the actual object #<MACRO expander> for the FENV.
                  (setq listr (cdr listr)) ; pop &WHOLE
                  (analyze1 (car listr) wholevar name wholevar)))
              (error-of-type 'source-program-error
+               :form %whole-form
+               :detail listr
                (TEXT "The lambda list of macro ~S contains an invalid &WHOLE: ~S")
                name listr)))
           ((eq item '&OPTIONAL)
@@ -320,12 +352,15 @@ the actual object #<MACRO expander> for the FENV.
              (cond ((symbolp item) (setq item (list item %default-form)))
                    ((and (consp item) (eql (length item) 1))
                     (setq item (list (car item) %default-form)))))
-           (cond ((symbolp item)
+           (cond ((null item) (empty-pattern name accessexp wholevar))
+                 ((symbolp item)
                   (setq %let-list (cons `(,item ,(cons-car accessexp))
                                         %let-list)))
                  ((atom item)
                   #1=
                   (error-of-type 'source-program-error
+                    :form %whole-form
+                    :detail item
                     (TEXT "The lambda list of macro ~S contains an invalid element ~S")
                     name item))
                  ((symbolp (car item))
@@ -357,7 +392,8 @@ the actual object #<MACRO expander> for the FENV.
           (t ; required arguments
            (setq %min-args (1+ %min-args))
            (setq %arg-count (1+ %arg-count))
-           (cond ((symbolp item)
+           (cond ((null item) (empty-pattern name accessexp wholevar))
+                 ((symbolp item)
                   (setq %let-list (cons `(,item ,(cons-car accessexp)) %let-list)))
                  ((atom item)
                   #1#) ; (error-of-type ... name item), s.o.
@@ -372,7 +408,9 @@ the actual object #<MACRO expander> for the FENV.
                     (rplaca (cdr (assoc g1 %let-list))
                             `(if ,test
                                  (error-of-type 'source-program-error
-                                   (TEXT "~S: ~S does not match lambda list element ~S")
+                                   :form ,wholevar
+                                   :detail ,g
+                                   (TEXT "~S: ~S does not match lambda list element ~:S")
                                    ',name ,g ',item)
                                  ,g)))))
            (setq accessexp (cons-cdr accessexp))))))
@@ -391,11 +429,13 @@ the actual object #<MACRO expander> for the FENV.
                 (nreconc l2 (cddr l1))))
            (cadr listr)))
         (error-of-type 'source-program-error
+          :form %whole-form
+          :detail lambdalist
           (TEXT "In the lambda list of macro ~S, &ENVIRONMENT must be followed by a non-NIL symbol: ~S")
           name lambdalist)))))
 
-(defun make-length-test (var &optional (header 1))
-  (let ((len `(LIST-LENGTH-DOTTED ,var)))
+(defun make-length-test (var &optional (header 1) (accept-dotted t))
+  (let ((len `(,(if accept-dotted 'LIST-LENGTH-DOTTED 'LIST-LENGTH) ,var)))
     (cond ((and (zerop %min-args) %restp) NIL)
           ((zerop %min-args) `(> ,len ,(+ header %arg-count)))
           (%restp `(< ,len ,(+ header %min-args)))
@@ -403,61 +443,100 @@ the actual object #<MACRO expander> for the FENV.
           (t `(NOT (<= ,(+ header %min-args)
                        ,len ,(+ header %arg-count)))))))
 
-(defun make-macro-expansion (macrodef &optional pre-process)
+(defun make-macro-expansion (macrodef whole-form
+                             ;; Optional arguments, for define-compiler-macro:
+                             &optional (valid-name-p #'symbolp)
+                                       pre-process)
   (when (atom macrodef)
     (error-of-type 'source-program-error
+      :form macrodef
+      :detail macrodef
       (TEXT "Cannot define a macro from that: ~S")
       macrodef))
-  (unless (symbolp (car macrodef))
+  (unless (funcall valid-name-p (car macrodef))
     (error-of-type 'source-program-error
+      :form macrodef
+      :detail (car macrodef)
       (TEXT "The name of a macro must be a symbol, not ~S")
       (car macrodef)))
   (when (atom (cdr macrodef))
     (error-of-type 'source-program-error
+      :form macrodef
+      :detail (cdr macrodef)
       (TEXT "Macro ~S is missing a lambda list.")
       (car macrodef)))
   (let ((name (car macrodef))
         (lambdalist (cadr macrodef))
         (body (cddr macrodef)))
     (multiple-value-bind (body-rest declarations docstring) (parse-body body t)
-      (if declarations (setq declarations
-                             (list (cons 'DECLARE declarations))))
-      (multiple-value-bind (newlambdalist envvar)
-          (remove-env-arg lambdalist name)
-        (let ((%arg-count 0) (%min-args 0) (%restp nil)
-              (%let-list nil) (%keyword-tests nil) (%default-form nil))
-          (analyze1 newlambdalist '(CDR <MACRO-FORM>) name '<MACRO-FORM>)
-          (let ((lengthtest (make-length-test '<MACRO-FORM>))
-                (mainform `(LET* ,(nreverse %let-list)
-                             ,@declarations
-                             ,@(nreverse %keyword-tests)
-                             (BLOCK ,name ,@body-rest))))
-            (if lengthtest
-              (setq mainform
-                    `(IF ,lengthtest
-                       (MACRO-CALL-ERROR <MACRO-FORM>)
-                       ,mainform)))
-            (values
-             `(FUNCTION ,name
-                (LAMBDA (<MACRO-FORM> ,(or envvar '<ENV-ARG>))
-                (DECLARE (CONS <MACRO-FORM>))
-                ,@(if envvar
-                    declarations ;; possibly contains a (declare (ignore envvar))
-                    '((DECLARE (IGNORE <ENV-ARG>))))
-                ,@(if docstring (list docstring))
-                ,@(if pre-process
-                    `((setq <MACRO-FORM> (,pre-process <MACRO-FORM>))))
-                ,mainform))
-             name
-             lambdalist
-             docstring)))))))
+      (if declarations
+        (setq declarations (list (cons 'DECLARE declarations))))
+      (let ((%whole-form whole-form))
+        (multiple-value-bind (newlambdalist envvar)
+            (remove-env-arg lambdalist name)
+          (let ((%arg-count 0) (%min-args 0) (%restp nil) (%null-tests nil)
+                (%let-list nil) (%keyword-tests nil) (%default-form nil))
+            (analyze1 newlambdalist '(CDR <MACRO-FORM>) name '<MACRO-FORM>)
+            (let ((lengthtest (make-length-test '<MACRO-FORM>))
+                  (mainform `(LET* ,(nreverse %let-list)
+                               ,@declarations
+                               ,@(nreverse %null-tests)
+                               ,@(nreverse %keyword-tests)
+                               (BLOCK ,(function-block-name name)
+                                 ,@body-rest))))
+              (if lengthtest
+                (setq mainform
+                      `(IF ,lengthtest
+                         (MACRO-CALL-ERROR <MACRO-FORM>)
+                         ,mainform)))
+              (let ((lambdabody
+                      `((<MACRO-FORM> ,(or envvar '<ENV-ARG>))
+                        (DECLARE (CONS <MACRO-FORM>))
+                        ,@(if envvar
+                            declarations ;; possibly contains a (declare (ignore envvar))
+                            '((DECLARE (IGNORE <ENV-ARG>))))
+                        ,@(if docstring (list docstring))
+                        ,@(if pre-process
+                            `((setq <MACRO-FORM> (,pre-process <MACRO-FORM>))))
+                        ,mainform)))
+                (values
+                  `(FUNCTION ,name (LAMBDA ,@lambdabody))
+                  lambdabody
+                  name
+                  lambdalist
+                  docstring)))))))))
 
-(defun make-macro-expander (macrodef &optional
-                            (env (vector (and (boundp '*venv*) *venv*)
-                                         (and (boundp '*fenv*) *fenv*)
-                                         (and (boundp '*benv*) *benv*)
-                                         (and (boundp '*genv*) *genv*)
-                                         (if (boundp '*denv*) *denv*
-                                             *toplevel-denv*))))
-  (make-macro ;; CLTL1: (eval (make-macro-expansion macrodef))
-   (evalhook (make-macro-expansion macrodef) nil nil env)))
+;; Creates a macro expander for MACROLET.
+(defun make-macro-expander (macrodef whole-form
+                            &optional
+                            ;; The environment is tricky: ANSI CL says that
+                            ;; 1. declarations, macros and symbol-macros from
+                            ;; outside can be used in the macro expander,
+                            ;; 2. other variable and function bindings cannot.
+                            ;; 3. It is unclear about BLOCK and TAGBODY tags.
+                            (env (vector (and (boundp '*venv*)
+                                              (cons 'MACROLET *venv*))
+                                         (and (boundp '*fenv*)
+                                              (cons 'MACROLET *fenv*))
+                                         (and (boundp '*benv*)
+                                              *benv*)
+                                         (and (boundp '*genv*)
+                                              *genv*)
+                                         (if (boundp '*denv*)
+                                           *denv*
+                                           *toplevel-denv*))))
+  (make-macro
+    (evalhook (make-macro-expansion macrodef whole-form) nil nil env)))
+
+;; Creates a macro expander for FUNCTION-MACRO-LET.
+;; Unlike in MACROLET, the macro is defined in the null lexical environment.
+(defun make-funmacro-full-lambdabody (macrodef)
+  (multiple-value-bind (expansion expansion-lambdabody)
+      (make-macro-expansion macrodef
+                            `(#:FUNCTION-MACRO-LET ,macrodef)) ; a dummy
+    (declare (ignore expansion))
+    expansion-lambdabody))
+(defun make-funmacro-expander (name full-lambdabody)
+  (make-macro
+    (evalhook `(FUNCTION ,name (LAMBDA ,@full-lambdabody)) nil nil
+              *toplevel-environment*)))
