@@ -43,21 +43,23 @@ to print the corresponding values, or T for all of them.")
   (or (eq t *apropos-do-more*) (memq what *apropos-do-more*)))
 
 (defun apropos (string &optional (package nil))
-  (dolist (sym (apropos-list string package) (terpri))
+  (dolist (sym (apropos-list string package))
     (format t "~&~s~40t" sym)
     (when (fboundp sym)
       (write-string "   ")
       (write-string (fbound-string sym))
       (when (apropos-do-more :function)
         (format t " [~s]" (fdefinition sym))))
+    (when (symbol-macro-expand sym)
+      (write-string "   ")
+      (write-string (TEXT "symbol-macro")))
     (when (boundp sym)
       (write-string "   ")
-      (cond ((constantp sym) (write-string (TEXT "constant")))
-            ((sys::symbol-macro-p (sys::%symbol-value sym))
-             (write-string (TEXT "symbol-macro")))
-            (t (write-string (TEXT "variable"))))
+      (if (constantp sym)
+        (write-string (TEXT "constant"))
+        (write-string (TEXT "variable")))
       (when (apropos-do-more :variable)
-        (format t " [~s]" (sys::%symbol-value sym))))
+        (format t " [~s]" (symbol-value sym))))
     (let ((type (or (get sym 'system::type-symbol)
                     (get sym 'system::defstruct-description))))
       (when type
@@ -70,7 +72,8 @@ to print the corresponding values, or T for all of them.")
         (write-string "   ")
         (write-string (TEXT "class"))
         (when (apropos-do-more :class)
-          (format t " [~s]" class)))))
+          (format t " [~s]" class))))
+    (elastic-newline))
   (values))
 
 ;;-----------------------------------------------------------------------------
@@ -79,64 +82,6 @@ to print the corresponding values, or T for all of them.")
 ;; Number of recursive calls since the top-level call.
 (defvar *describe-nesting*)
 
-;; A stream whose purpose is
-;; 1. to provide automatic indentation,
-;; 2. to distinguish top-level calls from recursive calls.
-(clos:defclass describe-stream (fundamental-character-output-stream)
-  ((target-stream :initarg :stream :type stream)
-   (buffer :type string :initform
-           (make-array (or *print-right-margin* sys::*prin-linelength*)
-                       :element-type 'character :fill-pointer 0
-                       :adjustable t))
-   (indent :initform 0 :type integer) ; current line's indentation
-   (pending-indent :initform nil :type (or null integer))))
-;; flush the buffer and print a newline (when NEWLINE-P is non-NIL)
-(defun describe-stream-flush-buffer (stream newline-p)
-  (clos:with-slots (target-stream buffer pending-indent indent) stream
-    (flet ((newline ()          ; terpri
-             (setq indent (* *describe-nesting* *print-indent-lists*)
-                   pending-indent indent)
-             (terpri target-stream)))
-      ;; fill: if the buffer does not fit on the line, TERPRI
-      (let ((pos (sys::line-position target-stream)))
-        (when (and pos
-                   (<= (or *print-right-margin* sys::*prin-linelength*)
-                       (+ (length buffer) pos)))
-          (newline)))
-      (when pending-indent      ; do the indent
-        (sys::write-spaces pending-indent target-stream)
-        (setq pending-indent nil))
-      (write-char-sequence buffer target-stream)
-      (setf (fill-pointer buffer) 0)
-      (when newline-p (newline)))))
-(clos:defmethod stream-write-char ((stream describe-stream) ch)
-  (clos:with-slots (target-stream buffer indent pending-indent) stream
-    (case ch
-      (#\Newline (describe-stream-flush-buffer stream t))
-      ((#\Space #\Tab)
-       (describe-stream-flush-buffer stream nil)
-       (write-char #\Space target-stream))
-      (t (vector-push-extend ch buffer)))))
-(clos:defmethod stream-line-column ((stream describe-stream))
-  (clos:with-slots (target-stream buffer indent) stream
-    (let ((pos (sys::line-position target-stream)))
-      (if pos (max (- (+ (length buffer) pos) indent) 0) nil))))
-(clos:defmethod stream-start-line-p ((stream describe-stream))
-  (clos:with-slots (target-stream buffer indent) stream
-    (let ((pos (sys::line-position target-stream)))
-      (if pos (<= (+ (length buffer) pos) indent) nil))))
-(clos:defmethod stream-finish-output ((stream describe-stream))
-  (describe-stream-flush-buffer stream nil)
-  (finish-output (clos:slot-value stream 'target-stream)))
-(clos:defmethod stream-force-output ((stream describe-stream))
-  (describe-stream-flush-buffer stream nil)
-  (force-output (clos:slot-value stream 'target-stream)))
-(clos:defmethod stream-clear-output ((stream describe-stream))
-  (clos:with-slots (target-stream buffer pending-indent) stream
-    (setq pending-indent nil)
-    (setf (fill-pointer buffer) 0)
-    (clear-output target-stream)))
-
 ; List of objects which have been described during the current top-level call.
 (defvar *describe-done*)
 
@@ -145,7 +90,8 @@ to print the corresponding values, or T for all of them.")
     (if slotnames
       (let* ((slotstrings (mapcar #'write-to-string slotnames)) more
              (tabpos (+ *print-indent-lists* 4 (reduce #'max (mapcar #'length slotstrings)))))
-        (format stream (TEXT "~%Slots:"))
+        (terpri stream)
+        (format stream (TEXT "Slots:"))
         (mapc #'(lambda (slotname slotstring)
                   (format stream "~%~V,0T  ~A~VT"
                           *print-indent-lists* slotstring tabpos)
@@ -155,7 +101,9 @@ to print the corresponding values, or T for all of them.")
                         ((format stream (TEXT "unbound")))))
               slotnames slotstrings)
         (dolist (vv (nreverse more)) (describe vv stream)))
-      (format stream (TEXT "~%No slots.")))))
+      (progn
+        (terpri stream)
+        (format stream (TEXT "No slots."))))))
 
 (clos:defgeneric describe-object (obj stream)
   (:method ((obj t) (stream stream))
@@ -184,7 +132,8 @@ to print the corresponding values, or T for all of them.")
        (format stream (TEXT "a symbol macro handler.")))
       (SYS::MACRO
        (format stream (TEXT "a macro expander."))
-       (format stream (TEXT "~%For more information, evaluate ~{~S~^ or ~}.")
+       (terpri stream)
+       (format stream (TEXT "For more information, evaluate ~{~S~^ or ~}.")
                `((DISASSEMBLE (MACRO-FUNCTION
                                ',(sys::closure-name
                                   (sys::macro-expander obj)))))))
@@ -200,6 +149,51 @@ to print the corresponding values, or T for all of them.")
                      value)
              (describe value stream))
            (format stream (TEXT "a GC-invisible pointer to a now defunct object.")))))
+      (EXT:WEAK-LIST
+       (let ((remaining (weak-list-list obj)))
+         (if remaining
+           (format stream (TEXT "a list of GC-invisible pointers to ~{~S~^, ~}.")
+                   remaining)
+           (format stream (TEXT "a list of GC-invisible pointers, all defunct by now.")))))
+      (EXT:WEAK-AND-RELATION
+       (let ((remaining (weak-and-relation-list obj)))
+         (if remaining
+           (format stream (TEXT "a weak \"and\" relation between ~{~S~^, ~}.")
+                   remaining)
+           (format stream (TEXT "a weak \"and\" relation, no longer referring to its objects.")))))
+      (EXT:WEAK-OR-RELATION
+       (let ((remaining (weak-or-relation-list obj)))
+         (if remaining
+           (format stream (TEXT "a weak \"or\" relation between ~{~S~^, ~}.")
+                   remaining)
+           (format stream (TEXT "a weak \"or\" relation, all elements defunct by now.")))))
+      (EXT:WEAK-MAPPING
+       (multiple-value-bind (key value alive) (weak-mapping-pair obj)
+         (if alive
+           (format stream (TEXT "a weak association from ~S to ~S.") key value)
+           (format stream (TEXT "a weak association, the key value being defunct by now.")))))
+      (EXT:WEAK-AND-MAPPING
+       (multiple-value-bind (keys value alive) (weak-and-mapping-pair obj)
+         (if alive
+           (format stream (TEXT "a weak \"and\" mapping from ~:S to ~S.") keys value)
+           (format stream (TEXT "a weak \"and\" mapping, some key value being defunct by now.")))))
+      (EXT:WEAK-OR-MAPPING
+       (multiple-value-bind (keys value alive) (weak-or-mapping-pair obj)
+         (if alive
+           (format stream (TEXT "a weak \"or\" mapping from ~:S to ~S.") keys value)
+           (format stream (TEXT "a weak \"or\" mapping, all keys being defunct by now.")))))
+      (EXT:WEAK-ALIST
+       (let ((type (weak-alist-type obj))
+             (remaining (weak-alist-contents obj)))
+         (format stream (TEXT "a weak association list, of type ~S ") type)
+         (ecase type
+           (:KEY    (format stream (TEXT "(i.e. a list of EXT:WEAK-MAPPING key/value pairs)")))
+           (:VALUE  (format stream (TEXT "(i.e. a list of EXT:WEAK-MAPPING value/key pairs)")))
+           (:KEY-AND-VALUE (format stream (TEXT "(i.e. a list of (key . value) pairs each combined into a EXT:WEAK-AND-RELATION)")))
+           (:KEY-OR-VALUE  (format stream (TEXT "(i.e. a list of (key . value) pairs each combined into a EXT:WEAK-OR-RELATION)"))))
+         (if remaining
+           (format stream (TEXT ", containing ~S.") remaining)
+           (format stream (TEXT ", no longer referring to any pairs.")))))
       (SYS::READ-LABEL
        (format stream (TEXT "a label used for resolving #~D# references during READ.")
                (logand (sys::address-of obj)
@@ -212,15 +206,16 @@ to print the corresponding values, or T for all of them.")
       (ADDRESS
        (format stream (TEXT "a machine address.")))))
   (:method ((obj clos:standard-object) (stream stream))
-      (format stream (TEXT "an instance of the CLOS class ~S.")
-              (clos:class-of obj))
-   (describe-slotted-object obj stream))
+    (format stream (TEXT "an instance of the CLOS class ~S.")
+            (clos:class-of obj))
+    (describe-slotted-object obj stream))
   (:method ((obj structure-object) (stream stream)) ; CLISP specific
     (format stream (TEXT "a structure of type ~S.")
             (type-of obj))
     (let ((types (butlast (cdr (sys::%record-ref obj 0)))))
       (when types
-        (format stream (TEXT "~%As such, it is also a structure of type ~{~S~^, ~}.")
+        (terpri stream)
+        (format stream (TEXT "As such, it is also a structure of type ~{~S~^, ~}.")
                 types)))
     (describe-slotted-object obj stream))
   (:method ((obj cons) (stream stream))
@@ -230,9 +225,10 @@ to print the corresponding values, or T for all of them.")
           (if (> len 1)
             (format stream (TEXT "a dotted list of length ~S.")
                     len)
-            (progn (format stream (TEXT "a cons."))
-                   (describe (car obj) stream)
-                   (describe (cdr obj) stream)))
+            (progn
+              (format stream (TEXT "a cons."))
+              (describe (car obj) stream)
+              (describe (cdr obj) stream)))
           (format stream (TEXT "a list of length ~S.")
                   len))
         (format stream (TEXT "a cyclic list.")))))
@@ -261,27 +257,26 @@ to print the corresponding values, or T for all of them.")
                    normal-printout)
                 (push pack accessible-packs)))))
         (when accessible-packs
-          (format stream (TEXT ", is accessible in ~:d package~:[~;s~] ~{~A~^, ~}")
-                  (length accessible-packs) (cdr accessible-packs)
+          (format stream (TEXT ", is accessible in ~:d package~:p ~{~A~^, ~}")
+                  (length accessible-packs)
                   (sort (mapcar #'package-name accessible-packs)
                         #'string<))))
       (when (keywordp obj)
         (format stream (TEXT ", is a keyword")))
+      (when (symbol-macro-expand obj)
+        (format stream (TEXT ", symbol-macro expanding to: ~S")
+                (macroexpand-1 obj))
+        (push `(macroexpand-1 ',obj) moree))
       (cond ((boundp obj)
              (if (constantp obj)
                (format stream (TEXT ", a constant"))
                (if (sys::special-variable-p obj)
                  (format stream (TEXT ", a variable declared SPECIAL"))
                  (format stream (TEXT ", a variable"))))
-             (when (symbol-macro-expand obj)
-               (format stream (TEXT " (macro: ~s)")
-                       (macroexpand-1 obj))
-               (push `(macroexpand-1 ',obj) moree))
-             (format stream (TEXT ", value: ~s") (sys::%symbol-value obj))
-             (pushnew (sys::%symbol-value obj) mored))
+             (format stream (TEXT ", value: ~s") (symbol-value obj))
+             (pushnew (symbol-value obj) mored))
             ((sys::special-variable-p obj)
-             (format stream
-                     (TEXT ", an unbound variable declared SPECIAL"))))
+             (format stream (TEXT ", an unbound variable declared SPECIAL"))))
       (when (fboundp obj)
         (format stream (TEXT ", names "))
         (cond ((special-operator-p obj)
@@ -290,10 +285,10 @@ to print the corresponding values, or T for all of them.")
                  (format stream (TEXT " with macro definition"))))
               ((functionp (symbol-function obj))
                (format stream (TEXT "a~:[~; deprecated~] function")
-                       (memq obj *deprecated-functions-list*)))
+                       (assoc obj *deprecated-functions-alist* :test #'eq)))
               (t ; (macro-function obj)
                (format stream (TEXT "a~:[~; deprecated~] macro")
-                       (memq obj *deprecated-functions-list*))))
+                       (assoc obj *deprecated-functions-alist* :test #'eq))))
         (let ((dep (get obj 'deprecated)))
           (when dep
             (format stream (TEXT " (use ~s instead)") dep)))
@@ -304,7 +299,7 @@ to print the corresponding values, or T for all of them.")
         (format stream (TEXT ", names a type"))
         (when (get obj 'system::deftype-expander)
           (push `(type-expand ',obj t) moree)))
-      (when (get obj 'clos::closclass)
+      (when (clos::defined-class-p (get obj 'clos::closclass))
         (format stream (TEXT ", names a class")))
       (when (symbol-plist obj)
         (let ((properties
@@ -318,14 +313,18 @@ to print the corresponding values, or T for all of them.")
       (dolist (ty '(compiler-macro setf structure type variable function))
         (let ((doc (documentation obj ty)))
           (when doc
-            (format stream (TEXT "~%Documentation as a ~a:~%~a") ty doc))))
+            (terpri stream)
+            (format stream (TEXT "Documentation as a ~a:") ty)
+            (terpri stream)
+            (princ doc stream))))
       (when moree
-        (format stream (TEXT "~%For more information, evaluate ~{~S~^ or ~}.")
+        (terpri stream)
+        (format stream (TEXT "For more information, evaluate ~{~S~^ or ~}.")
                 moree))
       (dolist (zz (nreverse mored)) (describe zz stream))))
   (:method ((obj integer) (stream stream))
-    (format stream (TEXT "an integer, uses ~S bit~:p, is represented as a ~(~A~).")
-            (integer-length obj) (type-of obj)))
+    (format stream (TEXT "an integer, uses ~S bit~:p, is represented as a ~:[bignum~;fixnum~].")
+            (integer-length obj) (sys::fixnump obj)))
   (:method ((obj ratio) (stream stream))
     (format stream (TEXT "a rational, not integral number.")))
   (:method ((obj float) (stream stream))
@@ -351,13 +350,16 @@ to print the corresponding values, or T for all of them.")
     (format stream (TEXT "."))
     #+UNICODE
     (let ((unicode-name (unicode-attributes obj)))
+      (terpri stream)
       (if unicode-name
-        (format stream (TEXT "~%Unicode name: ~A") unicode-name)
-        (format stream (TEXT "~%It is not defined by the Unicode standard."))))
-    (format stream (TEXT "~%It is a ~:[non-~;~]printable character.")
+        (format stream (TEXT "Unicode name: ~A") unicode-name)
+        (format stream (TEXT "It is not defined by the Unicode standard."))))
+    (terpri stream)
+    (format stream (TEXT "It is a ~:[non-~;~]printable character.")
             (graphic-char-p obj))
     (unless (standard-char-p obj)
-      (format stream (TEXT "~%Its use is non-portable."))))
+      (terpri stream)
+      (format stream (TEXT "Its use is non-portable."))))
   (:method ((obj stream) (stream stream))
     (format stream (TEXT "a~:[~:[ closed ~;n output-~]~;~:[n input-~;n input/output-~]~]stream.")
             (and (input-stream-p obj) (open-stream-p obj))
@@ -369,35 +371,57 @@ to print the corresponding values, or T for all of them.")
                 (package-name obj))
         (let ((nicknames (package-nicknames obj)))
           (when nicknames
-            (format stream (TEXT ". It has ~:d nickname~:[~;s~] ~{~A~^, ~}")
-                    (length nicknames) (cdr nicknames) nicknames)))
+            (format stream (TEXT ". It has ~:d nickname~:p ~{~A~^, ~}")
+                    (length nicknames) nicknames)))
         (format stream (TEXT "."))
         (let ((use-list (package-use-list obj))
               (used-by-list (package-used-by-list obj)))
-          (format stream (TEXT "~%It "))
+          (terpri stream)
+          (format stream (TEXT "It "))
           (when use-list
-            (format stream (TEXT "imports the external symbols of ~:d package~:[~;s~] ~{~A~^, ~} and ")
-                    (length use-list) (cdr use-list)
+            (format stream (TEXT "imports the external symbols of ~:d package~:p ~{~A~^, ~} and ")
+                    (length use-list)
                     (mapcar #'package-name use-list)))
           (let ((L nil) (count 0)) ; maybe list all exported symbols
             (do-external-symbols (s obj) (push s L) (incf count))
             (format stream
-                    (TEXT "exports ~[no symbols~:;~:*~:d symbol~:[~;s~]~]")
-                    count (/= 1 count))
+                    (TEXT "exports ~[no symbols~:;~:*~:d symbol~:p~]")
+                    count)
             (when (zerop *describe-nesting*)
               (format stream (TEXT "~{ ~S~^,~}")
                       (sort L #'string< :key #'symbol-name))))
           (if used-by-list
-            (format stream (TEXT " to ~:d package~:[~;s~] ~{~A~^, ~}")
-                    (length used-by-list) (cdr used-by-list)
+            (format stream (TEXT " to ~:d package~:p ~{~A~^, ~}")
+                    (length used-by-list)
                     (mapcar #'package-name used-by-list))
             (format stream (TEXT ", but no package uses these exports")))
-          (format stream (TEXT "."))))
+          (format stream (TEXT "."))
+          (let ((case-inverted-p (package-case-inverted-p obj))
+                (case-sensitive-p (package-case-sensitive-p obj))
+                (uses-cl (memq (find-package "COMMON-LISP") use-list))
+                (uses-cs-cl (memq (find-package "CS-COMMON-LISP") use-list)))
+            (cond ((and (not case-inverted-p) (not case-sensitive-p)
+                        (not uses-cs-cl)))
+                  ((and case-inverted-p case-sensitive-p (not uses-cl))
+                   (terpri stream)
+                   (format stream (TEXT "It is a modern case-sensitive package.")))
+                  (t
+                   (terpri stream)
+                   (format stream (TEXT "ATTENTION: "))
+                   (if case-inverted-p
+                     (if case-sensitive-p
+                       (format stream (TEXT "It is a modern case-sensitive package, but uses the symbols from the traditional ~S!")
+                               (find-package "COMMON-LISP"))
+                       (format stream (TEXT "It is case-inverted, but not case-sensitive!")))
+                     (if case-sensitive-p
+                       (format stream (TEXT "It is case-sensitive, but not case-inverted!"))
+                       (format stream (TEXT "It is a traditional ANSI CL compatible package, but uses the symbols from the modern ~S!")
+                               (find-package "CS-COMMON-LISP")))))))))
       (format stream (TEXT "a deleted package."))))
   (:method ((obj hash-table) (stream stream))
     (let ((count (hash-table-count obj)))
-      (format stream (TEXT "an ~s hash table with ~[no entries~:;~:*~:d entr~:[y~;ies~]~].")
-              (hash-table-test obj) count (/= 1 count))))
+      (format stream (TEXT "an ~s hash table with ~[no entries~:;~:*~:d entr~:*~[~;y~:;ies~]~].")
+              (hash-table-test obj) count)))
   (:method ((obj readtable) (stream stream))
     (format stream (TEXT "~:[a~;the Common Lisp~] readtable.")
             (equalp obj (copy-readtable))))
@@ -405,10 +429,9 @@ to print the corresponding values, or T for all of them.")
     (format stream (TEXT "a ~:[~;portable ~]pathname~:[.~;~:*, with the following components:~{~A~}~]")
             (sys::logical-pathname-p obj)
             (mapcan #'(lambda (kw component)
-                        (when component
-                          (list (format nil "~%~A = ~A"
-                                        (symbol-name kw)
-                                        (make-pathname kw component)))))
+                        (case component
+                          ((nil :unspecific)) ; ignore
+                          (t (list (format nil "~%~A = ~S" kw component)))))
                     '(:host :device :directory :name :type :version)
                     (list (pathname-host obj)
                           (pathname-device obj)
@@ -437,17 +460,41 @@ to print the corresponding values, or T for all of them.")
         (when (array-has-fill-pointer-p obj)
           (format stream (TEXT " and current length (fill-pointer) ~S")
                   (fill-pointer obj))))
-      (when (stringp obj)
+      (when (and (stringp obj) (not (eq eltype 'NIL)))
         #-UNICODE
         (format stream (TEXT " (a string)"))
         #+UNICODE
         (multiple-value-bind (bits ro-p realloc) (sys::string-info obj)
-          (write-string (TEXT " (a ") stream)
-          (when ro-p (write-string (TEXT "immutable ") stream))
-          (when realloc (write-string (TEXT "reallocated ") stream))
-          (format stream (TEXT "~A string)")
-                         (case bits (8 "ISO-8859-1") (16 "UCS-2") (32 "UCS-4")))))
+          (format stream (TEXT " (a~:[~;n immutable~] ~:[~;reallocated ~]~A string)")
+                  ro-p realloc
+                  (case bits (8 "ISO-8859-1") (16 "UCS-2") (32 "UCS-4")))))
       (format stream (TEXT "."))))
+  (:method ((obj generic-function) (stream stream))
+    (format stream (TEXT "a generic function."))
+    (unless (clos::generic-function-undeterminedp obj)
+      (terpri stream)
+      (format stream (TEXT "Argument list: ~A")
+              (clos:generic-function-lambda-list obj)))
+    (let ((mc (clos::method-combination-name (clos:generic-function-method-combination obj))))
+      (unless (eq mc 'STANDARD)
+        (terpri stream)
+        (format stream (TEXT "Method combination: ~S") mc)))
+    (let ((methods (clos:generic-function-methods obj)))
+      (terpri stream)
+      (if methods
+        (progn
+          (format stream (TEXT "Methods:"))
+          (dolist (meth (clos:generic-function-methods obj))
+            (format stream "~%  ~{~S ~}~S" (clos:method-qualifiers meth)
+                    (mapcar #'(lambda (specializer)
+                                (setq specializer (clos::specializer-pretty specializer))
+                                (if (and (clos::defined-class-p specializer)
+                                         (eq (find-class (class-name specializer) nil)
+                                             specializer))
+                                  (class-name specializer)
+                                  specializer))
+                            (clos:method-specializers meth)))))
+        (format stream (TEXT "No methods.")))))
   (:method ((obj function) (stream stream))
     (ecase (type-of obj)
       #+FFI
@@ -457,52 +504,65 @@ to print the corresponding values, or T for all of them.")
                                             (sys::%record-ref obj 2)
                                             (sys::%record-ref obj 3)
                                             (sys::%record-ref obj 4)))))
-      (COMPILED-FUNCTION ; SUBR
-       (format stream (TEXT "a built-in system function."))
+      (COMPILED-FUNCTION
        (multiple-value-bind (name req opt rest-p keywords other-keys)
            (sys::subr-info obj)
-         (when name
-           (sys::describe-signature stream req opt rest-p
-                                    keywords keywords other-keys))))
+         (if (and name req)
+           (progn
+             (format stream (TEXT "a built-in system function."))
+             (sys::describe-signature stream req opt rest-p
+                                      keywords keywords other-keys))
+           (progn
+             (format stream (TEXT "a compiled function."))
+             (multiple-value-bind (req opt rest-p key-p keywords other-keys-p)
+                 (sys::signature obj)
+               (sys::describe-signature stream req opt rest-p key-p keywords
+                                        other-keys-p)
+               (let* ((name (sys::closure-name obj))
+                      (funform (cond ((and (symbolp name) (macro-function name))
+                                      `(MACRO-FUNCTION ',name))
+                                     ((fboundp name) `(FUNCTION ,name)))))
+                 (when funform
+                   (terpri stream)
+                   (format stream
+                           (TEXT "For more information, evaluate ~{~S~^ or ~}.")
+                           `((DISASSEMBLE ,funform))))))))))
       (FUNCTION
-       (format stream
-               (TEXT "a~:[n interpret~; compil~]ed function.")
-               (compiled-function-p obj))
-       (if (compiled-function-p obj)
-         (multiple-value-bind (req opt rest-p key-p keywords other-keys-p)
-             (sys::signature obj)
-           (sys::describe-signature stream req opt rest-p key-p keywords
-                                    other-keys-p)
-           (let* ((name (sys::closure-name obj))
-                  (funform (cond ((and (symbolp name) (macro-function name))
-                                  `(MACRO-FUNCTION ',name))
-                                 ((fboundp name) `(FUNCTION ,name)))))
-             (when funform
-               (format stream
-                       (TEXT "~%For more information, evaluate ~{~S~^ or ~}.")
-                       `((DISASSEMBLE ,funform))))))
-         (let ((doc (sys::%record-ref obj 2)))
-           (format stream (TEXT "~%argument list: ~:S")
-                   (car (sys::%record-ref obj 1)))
-           (when doc
-             (format stream (TEXT "~%documentation: ~A") doc))))))))
+       (format stream (TEXT "an interpreted function."))
+       (let ((doc (sys::%record-ref obj 2)))
+         (terpri stream)
+         (format stream (TEXT "Argument list: ~:S")
+                 (car (sys::%record-ref obj 1)))
+         (when doc
+           (terpri stream)
+           (format stream (TEXT "Documentation: ~A") doc)))))))
 
 (defun describe1 (obj stream)
   (let ((objstring (sys::write-to-short-string
                     obj (or *print-right-margin* sys::*prin-linelength*))))
+    (fresh-line stream)
+    (terpri stream) ; blank line
     (if (memq obj *describe-done*)
-      (format stream (TEXT "~&~%~A [see above]") objstring)
-      (let ((doc (gethash obj *documentation*)))
+      (format stream (TEXT "~A [see above]") objstring)
+      (let ((doc (and (symbolp obj) (get obj 'sys::doc))))
         (push obj *describe-done*)
-        (format stream (TEXT "~&~%~A is ") objstring)
+        (format stream (TEXT "~A is ") objstring)
         (describe-object obj stream)
+        (fresh-line stream)
         (when doc
-          (format stream (TEXT "~&Documentation:"))
+          (terpri stream) ; blank line
+          (format stream (TEXT "Documentation:"))
           (do ((tail doc (cddr tail)))
               ((endp tail))
-            (format stream "~&~s:~%~s" (car tail) (cadr tail)))))))
+            (format stream "~&~s:~%~s" (car tail) (cadr tail)))
+          (fresh-line stream)))))
   (finish-output stream))
 
+;; A private class through which we can distinguish recursive describe calls
+;; from top-level calls.
+(defclass describe-stream (fill-stream)
+  ())
+
 (defun describe (obj &optional stream)
   (cond ((typep stream 'describe-stream) ; Recursive call
          ;; flush the pending output _before_ increasing indentation
@@ -520,23 +580,28 @@ to print the corresponding values, or T for all of them.")
                (*describe-nesting* 0)
                (*describe-done* nil))
            (describe1 obj (clos:make-instance 'describe-stream
-                                              :stream stream)))))
+                           :stream stream
+                           :text-indent '*describe-nesting*)))))
   (values))
 
 ;;-----------------------------------------------------------------------------
 ;; auxiliary functions for DESCRIBE of FUNCTION
 
-(defun arglist (func) (sig-to-list (get-signature func)))
+(defun arglist (func)
+  (setq func (coerce func 'function))
+  (if (typep func 'generic-function)
+    ; Generic functions store the lambda-list. It has meaningful variable names.
+    (clos:generic-function-lambda-list func)
+    ; Normal functions store only the signature, no variable names.
+    (sig-to-list (get-signature func))))
 
-(defun describe-signature (s req-anz opt-anz rest-p keyword-p keywords
+(defun describe-signature (stream req-anz opt-anz rest-p keyword-p keywords
                            allow-other-keys)
-  (when s
-    (format s (TEXT "~%Argument list: ")))
-  (prog1
-      (format s "(~{~A~^ ~})"
-              (signature-to-list req-anz opt-anz rest-p keyword-p keywords
-                                 allow-other-keys))
-    (when s (format s "."))))
+  (terpri stream)
+  (format stream (TEXT "Argument list: ~A.")
+          (format nil "(~{~A~^ ~})"
+                  (signature-to-list req-anz opt-anz rest-p keyword-p keywords
+                                     allow-other-keys))))
 
 ;;-----------------------------------------------------------------------------
 ;; auxiliary functions for CLISP metadata
@@ -694,42 +759,3 @@ to print the corresponding values, or T for all of them.")
 
 ;; DOCUMENTATION mit abfragen und ausgeben??
 ;; function, variable, type, structure, setf
-
-; Gibt object in einen String aus, der nach Möglichkeit höchstens max Spalten
-; lang sein soll.
-(defun write-to-short-string (object max)
-  ; Methode: probiere
-  ; level = 0: length = 0,1,2
-  ; level = 1: length = 1,2,3,4
-  ; level = 2: length = 2,...,6
-  ; usw. bis maximal level = 16.
-  ; Dabei level möglichst groß, und bei festem level length möglichst groß.
-  (if (or (numberp object) (symbolp object)) ; von length und level unbeeinflusst?
-    (write-to-string object)
-    (macrolet ((minlength (level) `,level)
-               (maxlength (level) `(* 2 (+ ,level 1))))
-      ; Um level möglist groß zu bekommen, dabei length = minlength wählen.
-      (let* ((level ; Binärsuche nach dem richtigen level
-               (let ((level1 0) (level2 16))
-                 (loop
-                   (when (= (- level2 level1) 1) (return))
-                   (let ((levelm (floor (+ level1 level2) 2)))
-                     (if (<= (string-width (write-to-string object :level levelm :length (minlength levelm))) max)
-                       (setq level1 levelm) ; levelm passt, probiere größere
-                       (setq level2 levelm) ; levelm passt nicht, probiere kleinere
-                 ) ) )
-                 level1
-             ) )
-             (length ; Binärsuche nach dem richtigen length
-               (let ((length1 (minlength level)) (length2 (maxlength level)))
-                 (loop
-                   (when (= (- length2 length1) 1) (return))
-                   (let ((lengthm (floor (+ length1 length2) 2)))
-                     (if (<= (string-width (write-to-string object :level level :length lengthm)) max)
-                       (setq length1 lengthm) ; lengthm passt, probiere größere
-                       (setq length2 lengthm) ; lengthm passt nicht, probiere kleinere
-                 ) ) )
-                 length1
-            )) )
-        (write-to-string object :level level :length length)
-) ) ) )
