@@ -19,7 +19,9 @@
   (and (symbolp obj)
        (gethash (symbol-name obj)
          (load-time-value
-           (make-hash-table :test #'equal
+           (make-hash-table
+             :key-type 'string :value-type 'symbol
+             :test 'fasthash-equal :warn-if-needs-rehash-after-gc t
              :initial-contents
                (mapcar #'(lambda (s) (cons (symbol-name s) s))
                  '(named
@@ -44,6 +46,8 @@
 ;; (loop-syntax-error loop-keyword) reports a syntax error.
 (defun loop-syntax-error (loop-keyword)
   (error-of-type 'source-program-error
+    :form *whole*
+    :detail loop-keyword        ; FIXME: should be something more useful
     (TEXT "~S: syntax error after ~A in ~S")
     'loop (symbol-name loop-keyword) *whole*))
 
@@ -184,8 +188,6 @@
                       ; so that these values
                       ; must be computed no later than in the initially-code.
 
-#+CLISP (remprop 'loop-initialization 'sys::defstruct-description)
-
 (proclaim '(inline li-vars))
 (defun li-vars (li)
   (case (li-specform li)
@@ -225,6 +227,8 @@
         (accunum-var nil) ; Akkumulationsvariable für count, sum etc.
         (accu-vars-nil nil) ; Akkumulationsvariablen mit Initialwert NIL
         (accu-vars-0 nil) ; Akkumulationsvariablen mit Initialwert 0
+        (accu-table (make-hash-table :warn-if-needs-rehash-after-gc t
+                                     :test 'stablehash-eq)) ; var --> clauses
         (accu-declarations nil) ; Typdeklarationen (umgedrehte Liste von declspecs)
         (initializations nil) ; Bindungen: (init ...) (umgedrehte Liste)
         (seen-for-as-= nil) ; schon eine FOR-AS-= Klausel gesehen?
@@ -243,7 +247,7 @@
              ((NCONC NCONCING APPEND APPENDING)
               (unless (eq (loop-keywordp (caddr rest)) 'INTO)
                 (return nil))))))
-        (results nil)) ; Liste von Ergebnisformen (höchstens eine!)
+        (results nil)) ; alist (value-form . (clause list))
     (labels
       ((next-kw () ; Schaut, ob als nächstes ein Keyword kommt.
                    ; Wenn ja, wird es geliefert. Wenn nein, Ergebnis NIL.
@@ -271,16 +275,43 @@
                    (CONS `(SETF ,tailvar (SETF ,accuvar ,incrementvar)))
                    (t `(SETF ,tailvar (LAST (SETF ,accuvar
                                                   ,incrementvar)))))))))
-       (acculist-var ()
+       (compatible-p (kw1 kw2)
+         ;; <http://www.lisp.org/HyperSpec/Body/sec_6-1-3.html>
+         (let ((ht #,(make-hash-table
+                      :warn-if-needs-rehash-after-gc t
+                      :test 'stablehash-eq
+                      :initial-contents
+                      '((collect . list) (append . list) (nconc . list)
+                        (sum . sum-count) (count . sum-count)
+                        (maximize . max-min) (minimize . max-min)))))
+           (eq (gethash kw1 ht) (gethash kw2 ht))))
+       (new-accu-var (var clause)
+         (let ((others (gethash var accu-table)) bad)
+           (when (setq bad (find (first clause) others
+                                 :key #'first :test-not #'compatible-p))
+             (error-of-type 'source-program-error
+               :form *whole* :detail clause
+               (TEXT "~S: variable ~S is used in incompatible clauses~{ ~A ~S~} and~{ ~A ~S~}")
+               *whole* var clause bad))
+           (setf (gethash var accu-table) (cons clause others))))
+       (new-result (form clause)
+         (let ((pair (assoc form results :test #'equal)))
+           (if pair
+               (push clause (cdr pair))
+               (push (list form clause) results))
+           results))
+       (acculist-var (keyword form)
          (or acculist-var
              (progn (setq acculist-var (gensym "ACCULIST-VAR-"))
                     (push acculist-var accu-vars-nil)
-                    (unless backward-consing-p
-                      (push acculist-var results))
+                    (let ((clause (list keyword form)))
+                      (new-accu-var acculist-var clause)
+                      (unless backward-consing-p
+                        (new-result acculist-var clause)))
                     acculist-var)))
-       (cons-backward (form) ; accuvar is NIL, accufuncsym is CONS
-         (let ((accuvar (acculist-var)))
-           (push `(SYS::LIST-NREVERSE ,accuvar) results)
+       (cons-backward (keyword form) ; accuvar is NIL, accufuncsym is CONS
+         (let ((accuvar (acculist-var keyword form)))
+           (new-result `(SYS::LIST-NREVERSE ,accuvar) `(,keyword ,form))
            `(SETQ ,accuvar (CONS ,form ,accuvar))))
        (parse-kw-p (kw) ; Schaut, ob als nächstes das Keyword kw kommt.
                         ; Wenn ja, wird es übergangen. Wenn nein, Ergebnis NIL.
@@ -301,6 +332,7 @@
          ;; return the variable pattern and the list of declspecs
          (unless (consp body-rest)
            (error-of-type 'source-program-error
+             :form *whole* :detail body-rest
              (TEXT "~S: missing variable.") 'loop))
          (let ((pattern (pop body-rest))
                (typedecl nil))
@@ -326,10 +358,12 @@
               (return))
             (push (pop body-rest) list))
            (nreverse list)))
-       (parse-nonempty-progn (kw) ;; after kw: parses {expr}* (CLtL2) or {expr}+ (CLHS)
+       (parse-nonempty-progn (kw) ;; after kw: [CLTS] {expr}+ or [CLTL2] {expr}*
          (let ((exprs (parse-progn)))
            (unless exprs
-             (warn (TEXT "~S: missing forms after ~A") 'loop (symbol-name kw)))
+             (if *loop-ansi*
+               (loop-syntax-error kw)
+               (warn (TEXT "~S: missing forms after ~A: permitted by CLtL2, forbidden by ANSI CL.") 'loop (symbol-name kw))))
            exprs))
        (parse-unconditional () ;; parse an unconditional
          ;; unconditional ::= {do | doing} {expr}*
@@ -387,22 +421,25 @@
                            ;; REVAPPEND/NRECONC now or before
                            (when backward-consing-p
                              (error "~s: internal error: backward consing should be illegal!" *whole*))
-                           (cons-forward form (acculist-var) accufuncsym))
+                           (cons-forward form (acculist-var kw form)
+                                         accufuncsym))
                           (t ; Unnamed acc var & CONS -> cons-backward
-                           (cons-backward form)))))
+                           (cons-backward kw form)))))
                  ((COUNT COUNTING SUM SUMMING MAXIMIZE MAXIMIZING
                    MINIMIZE MINIMIZING)
                   (pop body-rest)
-                  (let ((form (parse-form-or-it kw)) (type 'fixnum)
-                        (accuvar nil))
+                  (let* ((form (parse-form-or-it kw)) (type 'fixnum)
+                         (accuvar nil) (clause (list kw form)))
                     (when (parse-kw-p 'into)
                       (unless (and (consp body-rest)
                                    (symbolp (setq accuvar (pop body-rest))))
                         (loop-syntax-error 'into)))
                     (unless accuvar
                       (setq accuvar
-                        (or accunum-var (setq accunum-var (gensym))))
-                      (push accuvar results))
+                            (or accunum-var
+                                (setq accunum-var (gensym "ACCUNUM-VAR-"))))
+                      (new-result accuvar clause))
+                    (new-accu-var accuvar clause)
                     (when (consp body-rest)
                       (let ((kw2 (loop-keywordp (first body-rest))))
                         (when (or (not kw2) (eq kw2 'of-type))
@@ -498,6 +535,7 @@
            (dolist (var (li-vars initialization))
              (when (memq var var-list)
                (error-of-type 'source-program-error
+                 :form *whole* :detail var
                  (TEXT "~S: duplicate iteration variable ~S") *whole* var))
              (push var var-list))
            (push initialization initializations)))
@@ -538,14 +576,14 @@
                            (WHILE `(UNLESS ,form (LOOP-FINISH)))
                            (UNTIL `(WHEN ,form (LOOP-FINISH)))
                            (ALWAYS
-                             (push 'T results)
+                             (new-result 'T (list kw form))
                              `(UNLESS ,form (RETURN-FROM ,block-name 'NIL)))
                            (NEVER
-                             (push 'T results)
+                             (new-result 'T (list kw form))
                              `(WHEN ,form (RETURN-FROM ,block-name 'NIL)))
                            (THEREIS
                              (let ((dummy (gensym)))
-                               (push 'NIL results)
+                               (new-result 'NIL (list kw form))
                                `(BLOCK ,dummy
                                   (RETURN-FROM ,block-name
                                     (OR ,form (RETURN-FROM ,dummy NIL)))))))
@@ -555,8 +593,13 @@
                  (push `(PROGN ,@(parse-nonempty-progn kw)) initially-code))
                 ((FINALLY)
                  (pop body-rest)
-                 (push (or (parse-unconditional)
-                           `(PROGN ,@(parse-nonempty-progn kw)))
+                 (push (let ((form (parse-unconditional)))
+                         (if form
+                           (if *loop-ansi*
+                             (loop-syntax-error 'FINALLY)
+                             (warn (TEXT "~S: loop keyword immediately after ~A: permitted by CLtL2, forbidden by ANSI CL.") 'loop (symbol-name kw)))
+                           (setq form `(PROGN ,@(parse-nonempty-progn kw))))
+                          form)
                        finally-code))
                 ((WITH FOR AS REPEAT)
                  (pop body-rest)
@@ -627,14 +670,13 @@
                                        (step-function-form '(FUNCTION CDR))
                                        (step-function-var nil))
                                    (when (parse-kw-p 'by)
-                                     (setq step-function-form (parse-form 'by)))
-                                   (unless (and (consp step-function-form)
-                                                (eq (first step-function-form) 'FUNCTION)
-                                                (consp (cdr step-function-form))
-                                                (null (cddr step-function-form))
-                                                (symbolp (second step-function-form)))
-                                     (setq step-function-var (gensym)))
-                                   (let ((var (gensym))) ; Hilfsvariable
+                                     (setq step-function-form (parse-form 'by))
+                                     (unless (and (function-form-p step-function-form)
+                                                  (function-name-p (second step-function-form)))
+                                       (setq step-function-var (gensym))))
+                                   (let ((var (if (and pattern (symbolp pattern)
+                                                       (eq preposition 'ON))
+                                                  pattern (gensym))))
                                      (push `(,var ,start-form) bindings)
                                      (when step-function-var
                                        (push `(,step-function-var ,step-function-form)
@@ -645,13 +687,14 @@
                                                     'ENDP 'ATOM)
                                                 ,var)
                                           (LOOP-FINISH))))
-                                     (note-initialization
-                                       (make-loop-init
+                                     (unless (eq var pattern)
+                                       (note-initialization
+                                        (make-loop-init
                                          :specform 'LET
                                          :bindings (destructure pattern (if (eq preposition 'IN) `(CAR ,var) var))
                                          :declspecs new-declspecs
                                          :everytime t
-                                         :requires-stepbefore seen-endtest))
+                                         :requires-stepbefore seen-endtest)))
                                      (push
                                        (list var
                                              (if step-function-var
@@ -864,6 +907,7 @@
                                            (if (or (eq step-start-p 'down) (eq step-end-p 'down))
                                              (if (or (eq step-start-p 'up) (eq step-end-p 'up))
                                                (error-of-type 'source-program-error
+                                                 :form *whole* :detail kw
                                                  (TEXT "~S: questionable iteration direction after ~A")
                                                  'loop (symbol-name kw))
                                                'down)
@@ -874,6 +918,7 @@
                                          ; Abwärtsiteration ohne Startwert ist nicht erlaubt.
                                          ; Die zweite optionale Klausel (d.h. preposition) muss abwärts zeigen.
                                          (error-of-type 'source-program-error
+                                           :form *whole* :detail preposition
                                            (TEXT "~S: specifying ~A requires FROM or DOWNFROM")
                                            'loop (symbol-name preposition)))
                                        ; Aufwärtsiteration -> Startwert 0
@@ -933,15 +978,17 @@
                       (note-initialization
                        (make-endtest `(UNLESS (PLUSP ,var) (LOOP-FINISH))))))))
                 (t (error-of-type 'source-program-error
+                     :form *whole* :detail *whole*
                      (TEXT "~S: illegal syntax near ~S in ~S")
                      'loop (first body-rest) *whole*)))))))
       ; Noch einige semantische Tests:
-      (setq results (delete-duplicates results :test #'equal))
       (when (> (length results) 1)
         (error-of-type 'source-program-error
-          (TEXT "~S: ambiguous result of loop ~S") 'loop *whole*))
+          :form *whole* :detail *whole*
+          (TEXT "~S: ambiguous result:~:{~%~S from ~@{~{~A ~S~}~^, ~}~}")
+          *whole* results))
       (unless (null results)
-        (push `(RETURN-FROM ,block-name ,@results) finally-code))
+        (push `(RETURN-FROM ,block-name ,(caar results)) finally-code))
       ; Initialisierungen abarbeiten und optimieren:
       (let ((initializations1
              (unless (zerop (length *helpvars*))
@@ -1009,6 +1056,7 @@
         (flet ((check-accu-var (var)
                  (when (memq var var-list)
                    (error-of-type 'source-program-error
+                     :form *whole* :detail var
                      (TEXT "~S: accumulation variable ~S is already bound")
                      *whole* var))))
           (push
@@ -1065,7 +1113,7 @@
 
 ;; Der eigentliche Macro:
 (defmacro loop (&whole whole &body body)
-  (if (some #'loop-keywordp body)
+  (if (some #'atom body)
     ;; "extended" loop form
     (expand-loop whole body)
     ;; "simple" loop form
